import typing

import python_dice.interface.i_probability_distribution as i_probability_distribution
import python_dice.interface.i_probability_state as i_probability_state


class ProbabilityState(i_probability_state.IProbabilityDistributionState):
    def get_var_dict(self):
        return self._var

    def get_constant_dict(self):
        return self._constants

    def __init__(self):
        self._var: typing.Dict[
            str, i_probability_distribution.IProbabilityDistribution
        ] = {}
        self._constants: typing.Dict[str, int] = {}

    def has_constant(self, name) -> bool:
        return name in self._constants

    def has_var(self, name: str) -> bool:
        return name in self._var

    def get_constant(self, name: str) -> int:
        if not self.has_constant(name):
            raise KeyError(f"Constant named {name} not in state.")
        return self._constants[name]

    def get_var(self, name: str) -> i_probability_distribution.IProbabilityDistribution:
        if not self.has_var(name):
            raise KeyError(f"Var named {name} not in state.")
        return self._var[name]

    def set_constant(self, name: str, value: int) -> None:
        self._constants[name] = value

    def set_var(
        self,
        name: str,
        distribution: i_probability_distribution.IProbabilityDistribution,
    ) -> None:
        self._var[name] = distribution



import typing

import rply

import python_dice.interface.i_probability_distribution as i_probability_distribution
import python_dice.interface.python_dice_expression.i_dice_expression as i_dice_expression
import python_dice.src.probability_distribution as probability_distribution


class ConstantIntegerExpression(i_dice_expression.IDiceExpression):
    TOKEN_RULE = """expression : CONSTANT_INTEGER"""

    @staticmethod
    def add_production_function(
        parser_generator: rply.ParserGenerator,
    ) -> typing.Callable:
        @parser_generator.production(ConstantIntegerExpression.TOKEN_RULE)
        def constant_integer(_, tokens) -> i_dice_expression.IDiceExpression:
            return ConstantIntegerExpression(tokens[0].value)

        return constant_integer

    def __init__(self, number: str):
        self._number = number

    def roll(self) -> int:
        return int(self._number)

    def max(self) -> int:
        return int(self._number)

    def min(self) -> int:
        return int(self._number)

    def __str__(self) -> str:
        return self._number

    def estimated_cost(self) -> int:
        return 2

    def get_probability_distribution(
        self,
    ) -> i_probability_distribution.IProbabilityDistribution:
        return probability_distribution.ProbabilityDistribution({int(self._number): 1})



import re
import typing


def _find_range(value: str) -> range:
    split_range = re.split("-", value)
    range_start: int
    range_end: int

    if re.match(r"\d+--?\d+", value) is not None:
        range_start = int(split_range[0])
    else:
        range_start = -int(split_range[1])
    if re.match(r"-?\d+--\d+", value) is not None:
        range_end = -int(split_range[-1])
    else:
        range_end = int(split_range[-1])
    return range(min(range_start, range_end), max(range_start, range_end) + 1)


def get_single_dice_outcome_map(dice_type_string: str) -> typing.Dict[int, int]:
    def safe_add(dictionary, value, amount):
        if value not in dictionary:
            dictionary[value] = 0
        dictionary[value] += amount

    if re.match(r"\d+", dice_type_string) is not None:
        value_dictionary = {value: 1 for value in range(1, int(dice_type_string) + 1)}
    elif dice_type_string == "%":
        value_dictionary = {value: 1 for value in range(1, 100 + 1)}
    elif dice_type_string == "F":
        value_dictionary = {-1: 1, 0: 1, 1: 1}
    else:
        value_dictionary = {}
        for value in re.split(r",", re.sub(r"(\[|\]|\s+)", "", dice_type_string)):
            multiplier = "1"
            if re.match(r"(-?\d+--?\d+|-?\d+)\*\d+", value) is not None:
                value, multiplier = re.split(r"\*", value)
            if re.match(r"-?\d+--?\d+", value) is not None:
                range_values = _find_range(value)
            else:
                range_values = [int(value)]
            for key in range_values:
                safe_add(value_dictionary, key, int(multiplier))
    return value_dictionary



import typing

import rply

import python_dice.interface.i_probability_state as i_probability_state
import python_dice.interface.i_python_dice_lexer as i_pydice_lexer
import python_dice.interface.i_python_dice_parser as i_pydice_parser
import python_dice.interface.python_dice_expression.i_dice_expression as i_dice_expression
import python_dice.src.probability_state as probability_state
import python_dice.src.python_dice_lexer as pydice_lexer
import python_dice.src.python_dice_syntax_objects as pydice_syntax_objects


class PythonDiceParser(i_pydice_parser.IPythonDiceParser):
    def __init__(self, lexer: i_pydice_lexer.IPythonDiceLexer = None):
        if lexer is None:
            lexer = pydice_lexer.PythonDiceLexer()

        self._lexer = lexer
        parser_generator = rply.ParserGenerator(
            [syntax.get_token_name() for syntax in pydice_syntax_objects.LEXER_SYNTAX],
            precedence=pydice_syntax_objects.PRECEDENCE,
        )

        for syntax in pydice_syntax_objects.PARSER_EXPRESSIONS:
            syntax.add_production_function(parser_generator)

        #  pylint: disable=unused-variable
        @parser_generator.error
        def error_handler(
            _: probability_state.ProbabilityState, token: rply.token.Token
        ):
            raise ValueError(
                "Ran into a %s (%s) where it wasn't expected, at position %s."
                % (token.gettokentype(), token.getstr(), token.getsourcepos())
            )

        self._parser = parser_generator.build()

    def parse(
        self,
        input_text: str,
        state: i_probability_state.IProbabilityDistributionState = None,
    ) -> typing.Tuple[
        i_dice_expression.IDiceExpression,
        i_probability_state.IProbabilityDistributionState,
    ]:
        if state is None:
            state = probability_state.ProbabilityState()
        return self._parser.parse(iter(self._lexer.lex(input_text)), state=state), state



import typing

import rply

import python_dice.interface.i_probability_distribution as i_probability_distribution
import python_dice.interface.i_probability_state as i_probability_state
import python_dice.interface.python_dice_expression.i_dice_expression as i_dice_expression


class GetVarExpression(i_dice_expression.IDiceExpression):
    TOKEN_RULE = """expression : NAME"""

    @staticmethod
    def add_production_function(
        parser_generator: rply.ParserGenerator,
    ) -> typing.Callable:
        @parser_generator.production(GetVarExpression.TOKEN_RULE)
        def get_var(state, tokens) -> i_dice_expression.IDiceExpression:
            return GetVarExpression(state, tokens[0].value)

        return get_var

    def __init__(
        self, state: i_probability_state.IProbabilityDistributionState, name: str
    ):
        self._state = state
        self._name = name

    def roll(self) -> int:
        return self._state.get_constant(self._name)

    def max(self) -> int:
        return self._state.get_constant(self._name)

    def min(self) -> int:
        return self._state.get_constant(self._name)

    def __str__(self) -> str:
        return f"{self._name}"

    def estimated_cost(self) -> int:
        return self._state.get_constant(self._name)

    def get_probability_distribution(
        self,
    ) -> i_probability_distribution.IProbabilityDistribution:
        return self._state.get_var(self._name)



import typing

import rply

import python_dice.interface.i_probability_distribution as i_probability_distribution
import python_dice.interface.python_dice_expression.i_dice_expression as i_dice_expression


class MinMaxExpression(i_dice_expression.IDiceExpression):

    TOKEN_RULE = """expression : MINMAX OPEN_PARENTHESIS expression COMMA expression CLOSE_PARENTHESIS"""
    OPERATOR_MAP = {"MAX": max, "MIN": min}

    @staticmethod
    def add_production_function(
        parser_generator: rply.ParserGenerator,
    ) -> typing.Callable:
        @parser_generator.production(MinMaxExpression.TOKEN_RULE)
        def min_max(_, tokens) -> i_dice_expression.IDiceExpression:
            return MinMaxExpression(tokens[0].value, tokens[2], tokens[4])

        return min_max

    def __init__(
        self,
        min_max: str,
        expression_one: i_dice_expression.IDiceExpression,
        expression_two: i_dice_expression.IDiceExpression,
    ):
        self._min_max = min_max
        self._expression_one = expression_one
        self._expression_two = expression_two

    def roll(self) -> int:
        return self.OPERATOR_MAP[self._min_max](
            self._expression_one.roll(), self._expression_two.roll()
        )

    def max(self) -> int:
        return self.OPERATOR_MAP[self._min_max](
            self._expression_one.max(), self._expression_two.max()
        )

    def min(self) -> int:
        return self.OPERATOR_MAP[self._min_max](
            self._expression_one.min(), self._expression_two.min()
        )

    def __str__(self) -> str:
        return (
            f"{self._min_max}({str(self._expression_one)}, {str(self._expression_two)})"
        )

    def estimated_cost(self) -> int:
        return (
            self._expression_one.estimated_cost()
            + self._expression_two.estimated_cost()
        )

    def get_probability_distribution(
        self,
    ) -> i_probability_distribution.IProbabilityDistribution:
        if self._min_max == "MAX":
            return self._expression_one.get_probability_distribution().max_operator(
                self._expression_two.get_probability_distribution()
            )
        return self._expression_one.get_probability_distribution().min_operator(
            self._expression_two.get_probability_distribution()
        )



import typing

import rply

import python_dice.interface.i_probability_distribution as i_probability_distribution
import python_dice.interface.python_dice_expression.i_dice_expression as i_dice_expression


class SubtractExpression(i_dice_expression.IDiceExpression):
    TOKEN_RULE = """expression : expression SUBTRACT expression"""

    @staticmethod
    def add_production_function(
        parser_generator: rply.ParserGenerator,
    ) -> typing.Callable:
        @parser_generator.production(SubtractExpression.TOKEN_RULE)
        def subtract(_, tokens) -> i_dice_expression.IDiceExpression:
            return SubtractExpression(tokens[0], tokens[2])

        return subtract

    def __init__(
        self,
        expression_one: i_dice_expression.IDiceExpression,
        expression_two: i_dice_expression.IDiceExpression,
    ):
        self._expression_one = expression_one
        self._expression_two = expression_two

    def roll(self) -> int:
        return self._expression_one.roll() - self._expression_two.roll()

    def max(self) -> int:
        return self.get_probability_distribution().max()

    def min(self) -> int:
        return self.get_probability_distribution().min()

    def __str__(self) -> str:
        return f"{str(self._expression_one)} - {str(self._expression_two)}"

    def estimated_cost(self) -> int:
        return (
            self._expression_one.estimated_cost()
            + self._expression_two.estimated_cost()
        )

    def get_probability_distribution(
        self,
    ) -> i_probability_distribution.IProbabilityDistribution:
        return (
            self._expression_one.get_probability_distribution()
            - self._expression_two.get_probability_distribution()
        )



import random
import typing

import rply

import python_dice.interface.i_probability_distribution as i_probability_distribution
import python_dice.interface.python_dice_expression.i_dice_expression as i_dice_expression
import python_dice.src.probability_distribution as probability_distribution
import python_dice.src.python_dice_expression.dice_expression_helper as dice_expression_helper


class DiceExpression(i_dice_expression.IDiceExpression):
    TOKEN_RULE = """expression : DICE"""

    @staticmethod
    def add_production_function(
        parser_generator: rply.ParserGenerator,
    ) -> typing.Callable:
        @parser_generator.production(DiceExpression.TOKEN_RULE)
        def dice(_, tokens) -> i_dice_expression.IDiceExpression:
            return DiceExpression(tokens[0].value)

        return dice

    def __init__(self, string_form: str):
        self._string_form = string_form
        self._single_dice_outcome_map = (
            dice_expression_helper.get_single_dice_outcome_map(
                self._string_form.split("d")[1]
            )
        )
        self._number_of_dice = self._get_number_of_dice()

    def _get_number_of_dice(self) -> int:
        string_num = self._string_form.split("d")[0]
        return 1 if string_num == "" else int(string_num)

    def roll(self) -> int:
        return sum(
            random.choices(
                list(self._single_dice_outcome_map.keys()),
                weights=list(self._single_dice_outcome_map.values()),
                k=self._number_of_dice,
            )
        )

    def max(self) -> int:
        return (max(self._single_dice_outcome_map.keys())) * self._number_of_dice

    def min(self) -> int:
        return (min(self._single_dice_outcome_map.keys())) * self._number_of_dice

    def estimated_cost(self) -> int:
        return self._number_of_dice * len(self._single_dice_outcome_map.values())

    def __str__(self) -> str:
        return f"{self._string_form}"

    def get_probability_distribution(
        self,
    ) -> i_probability_distribution.IProbabilityDistribution:
        single_dice_distribution = probability_distribution.ProbabilityDistribution(
            self._single_dice_outcome_map
        )
        distribution = probability_distribution.ProbabilityDistribution(
            single_dice_distribution.get_result_map()
        )
        for _ in range(0, self._number_of_dice - 1):
            distribution = distribution + single_dice_distribution
        return distribution



import python_dice.interface.python_dice_syntax.i_dice_syntax as i_dice_statement
import python_dice.src.python_dice_syntax.dice_syntax_helper as dice_syntax_helper


class DiceSyntax(i_dice_statement.IDiceSyntax):
    TOKEN_NAME = "DICE"
    TOKEN_REGEX = dice_syntax_helper.DICE_SYNTAX

    @staticmethod
    def get_token_name() -> str:
        return DiceSyntax.TOKEN_NAME

    @staticmethod
    def get_token_regex() -> str:
        return DiceSyntax.TOKEN_REGEX



import python_dice.src.python_dice_expression.abs_expression as abs_expression
import python_dice.src.python_dice_expression.add_expression as add_expression
import python_dice.src.python_dice_expression.binary_operator_expression as binary_operator_expression
import python_dice.src.python_dice_expression.constant_binary_expression as constant_binary_expression
import python_dice.src.python_dice_expression.constant_integer_expression as constant_integer_expression
import python_dice.src.python_dice_expression.dice_expression as dice_expression
import python_dice.src.python_dice_expression.drop_keep_expression as drop_keep_expression
import python_dice.src.python_dice_expression.get_var_expression as get_var_expression
import python_dice.src.python_dice_expression.integer_division_expression as integer_division_expression
import python_dice.src.python_dice_expression.minmax_expression as minmax_expression
import python_dice.src.python_dice_expression.multiply_expression as multiply_expression
import python_dice.src.python_dice_expression.not_expression as not_expression
import python_dice.src.python_dice_expression.parentheses_enclosed_expression as parentheses_enclosed_expression
import python_dice.src.python_dice_expression.subtract_expression as subtract_expression
import python_dice.src.python_dice_expression.var_assignment_expression as var_assignment_expression
import python_dice.src.python_dice_syntax.abs_syntax as abs_syntax
import python_dice.src.python_dice_syntax.add_syntax as add_syntax
import python_dice.src.python_dice_syntax.assignment_syntax as assignment_syntax
import python_dice.src.python_dice_syntax.binary_operator_syntax as binary_operator_syntax
import python_dice.src.python_dice_syntax.close_parenthesis_syntax as close_parenthesis_syntax
import python_dice.src.python_dice_syntax.comma_syntax as comma_syntax
import python_dice.src.python_dice_syntax.constant_binary_syntax as constant_binary_syntax
import python_dice.src.python_dice_syntax.constant_integer_syntax as constant_integer_syntax
import python_dice.src.python_dice_syntax.dice_syntax as dice_syntax
import python_dice.src.python_dice_syntax.drop_keep_syntax as drop_keep_syntax
import python_dice.src.python_dice_syntax.integer_division_syntax as integer_division_syntax
import python_dice.src.python_dice_syntax.min_max_syntax as min_max_syntax
import python_dice.src.python_dice_syntax.multiply_syntax as multiply_syntax
import python_dice.src.python_dice_syntax.name_syntax as name_syntax
import python_dice.src.python_dice_syntax.not_syntax as not_syntax
import python_dice.src.python_dice_syntax.open_parenthesis_syntax as open_parenthesis_syntax
import python_dice.src.python_dice_syntax.subtract_syntax as subtract_syntax
import python_dice.src.python_dice_syntax.var_syntax as var_syntax

LEXER_SYNTAX = [
    open_parenthesis_syntax.OpenParenthesisSyntax,
    close_parenthesis_syntax.CloseParenthesisSyntax,
    comma_syntax.CommaSyntax,
    not_syntax.NotSyntax,
    drop_keep_syntax.DropKeepSyntax,
    dice_syntax.DiceSyntax,
    binary_operator_syntax.BinaryOperatorSyntax,
    constant_integer_syntax.ConstantIntegerSyntax,
    constant_binary_syntax.ConstantBinarySyntax,
    add_syntax.AddSyntax,
    subtract_syntax.SubtractSyntax,
    multiply_syntax.MultiplySyntax,
    integer_division_syntax.IntegerDivisionSyntax,
    min_max_syntax.MinMaxSyntax,
    abs_syntax.AbsSyntax,
    assignment_syntax.AssignmentSyntax,
    var_syntax.VarSyntax,
    name_syntax.NameSyntax,
]

PARSER_EXPRESSIONS = [
    parentheses_enclosed_expression.ParenthesisEnclosedExpression,
    not_expression.NotExpression,
    drop_keep_expression.DropKeepExpression,
    dice_expression.DiceExpression,
    constant_integer_expression.ConstantIntegerExpression,
    constant_binary_expression.ConstantBinaryExpression,
    add_expression.AddExpression,
    subtract_expression.SubtractExpression,
    multiply_expression.MultiplyExpression,
    integer_division_expression.IntegerDivisionExpression,
    binary_operator_expression.BinaryOperatorExpression,
    minmax_expression.MinMaxExpression,
    abs_expression.AbsExpression,
    var_assignment_expression.VarAssignmentExpression,
    get_var_expression.GetVarExpression,
]

PRECEDENCE = [
    ("nonassoc", [var_syntax.VarSyntax.get_token_name()]),
    ("left", [assignment_syntax.AssignmentSyntax.get_token_name()]),
    ("left", [open_parenthesis_syntax.OpenParenthesisSyntax.get_token_name()]),
    ("left", [not_syntax.NotSyntax.get_token_name()]),
    (
        "left",
        [
            add_syntax.AddSyntax.get_token_name(),
            subtract_syntax.SubtractSyntax.get_token_name(),
        ],
    ),
    (
        "left",
        [
            integer_division_syntax.IntegerDivisionSyntax.get_token_name(),
            multiply_syntax.MultiplySyntax.get_token_name(),
        ],
    ),
    ("left", [binary_operator_syntax.BinaryOperatorSyntax.get_token_name()]),
]



import abc

import rply

import python_dice.src.probability_distribution as probability_distribution


class IDiceExpression(abc.ABC, rply.token.Token):
    @staticmethod
    @abc.abstractmethod
    def add_production_function(parser_generator: rply.ParserGenerator) -> None:
        """

        add a production rule to the parser generator
        """

    @abc.abstractmethod
    def roll(self) -> int:
        """

        :return: returns the result of a roll of the dice
                 <int>
        """

    @abc.abstractmethod
    def max(self) -> int:
        """

        :return: returns a largest possible result of a roll of the dice
                 <int>
        """

    @abc.abstractmethod
    def min(self) -> int:
        """

        :return: returns a smallest possible result of a roll of the dice
                 <int>
        """

    @abc.abstractmethod
    def estimated_cost(self) -> int:
        """

        :return: returns a estimated cost of the python dice expression
                 <int>
        """

    @abc.abstractmethod
    def get_probability_distribution(
        self,
    ) -> probability_distribution.ProbabilityDistribution:
        """

        :return: the probability of the statement
        """



import abc

import python_dice.interface.i_probability_distribution as i_probability_distribution


class IProbabilityDistributionState(abc.ABC):
    @abc.abstractmethod
    def get_var(self, name: str) -> i_probability_distribution.IProbabilityDistribution:
        pass

    @abc.abstractmethod
    def has_var(self, name: str) -> bool:
        pass

    @abc.abstractmethod
    def set_var(
        self,
        name: str,
        distribution: i_probability_distribution.IProbabilityDistribution,
    ) -> None:
        pass

    @abc.abstractmethod
    def get_var_dict(self):
        pass

    @abc.abstractmethod
    def get_constant(self, name: str) -> int:
        pass

    @abc.abstractmethod
    def has_constant(self, name: str) -> bool:
        pass

    @abc.abstractmethod
    def set_constant(self, name: str, value: int) -> None:
        pass

    @abc.abstractmethod
    def get_constant_dict(self):
        pass



import random
import re
import typing

import rply

import python_dice.interface.i_probability_distribution as i_probability_distribution
import python_dice.interface.python_dice_expression.i_dice_expression as i_dice_expression
import python_dice.src.probability_distribution as probability_distribution
import python_dice.src.python_dice_expression.constant_integer_expression as constant_integer_expression
import python_dice.src.python_dice_expression.dice_expression as dice_expression
import python_dice.src.python_dice_expression.dice_expression_helper as dice_expression_helper


class DropKeepExpression(i_dice_expression.IDiceExpression):
    TOKEN_RULE = """expression : DROP_KEEP_DICE"""

    @staticmethod
    def add_production_function(
        parser_generator: rply.ParserGenerator,
    ) -> typing.Callable:
        @parser_generator.production(DropKeepExpression.TOKEN_RULE)
        def drop_keep(_, tokens) -> i_dice_expression.IDiceExpression:
            return DropKeepExpression(tokens[0].value)

        return drop_keep

    def __init__(self, string_form: str):
        self._string_form = string_form
        self._number_of_dice = self._get_number_of_dice()
        self._single_dice_outcome_map = (
            dice_expression_helper.get_single_dice_outcome_map(
                re.split(r"[dk]", self._string_form)[1]
            )
        )
        self._number_to_keep_or_drop = self._get_number_to_keep_or_drop()
        self._simplified_form = None

        if self._is_keep():
            if self._number_of_dice <= self._number_to_keep_or_drop:
                self._simplified_form = dice_expression.DiceExpression(
                    "%dd%s" % (self._number_of_dice, self._get_number_of_sides_string())
                )
            elif self._number_to_keep_or_drop == 0:
                self._simplified_form = (
                    constant_integer_expression.ConstantIntegerExpression("0")
                )
        else:
            if self._number_of_dice <= self._number_to_keep_or_drop:
                self._simplified_form = (
                    constant_integer_expression.ConstantIntegerExpression("0")
                )
            elif self._number_to_keep_or_drop == 0:
                self._simplified_form = dice_expression.DiceExpression(
                    "%dd%s" % (self._number_of_dice, self._get_number_of_sides_string())
                )

    def _get_number_of_dice(self) -> int:
        string_num = re.split(r"[dk]", self._string_form)[0]
        return 1 if string_num == "" else int(string_num)

    def _get_number_to_keep_or_drop(self) -> int:
        return int(re.split(r"[dk]", self._string_form)[2])

    def _get_number_of_sides_string(self) -> str:
        return re.split(r"[dk]", self._string_form)[1]

    def _is_keep(self) -> bool:
        return "k" in self._string_form

    def roll(self) -> int:
        if self._simplified_form is not None:
            return self._simplified_form.roll()

        dice_rolls = random.choices(
            list(self._single_dice_outcome_map.keys()),
            weights=list(self._single_dice_outcome_map.values()),
            k=self._number_of_dice,
        )
        dice_rolls.sort()
        if self._is_keep():
            return sum(dice_rolls[-self._number_to_keep_or_drop :])
        return sum(dice_rolls[: self._number_of_dice - self._number_to_keep_or_drop])

    def max(self) -> int:
        if self._simplified_form is not None:
            return self._simplified_form.max()

        if self._is_keep():
            return self._number_to_keep_or_drop * max(
                self._single_dice_outcome_map.keys()
            )
        return (self._number_of_dice - self._number_to_keep_or_drop) * max(
            self._single_dice_outcome_map.keys()
        )

    def min(self) -> int:
        if self._simplified_form is not None:
            return self._simplified_form.min()

        if self._is_keep():
            return self._number_to_keep_or_drop * min(
                self._single_dice_outcome_map.keys()
            )
        return (self._number_of_dice - self._number_to_keep_or_drop) * min(
            self._single_dice_outcome_map.keys()
        )

    def __str__(self) -> str:
        return f"{self._string_form}"

    def estimated_cost(self) -> int:
        if self._simplified_form is not None:
            return self._simplified_form.estimated_cost()

        if self._is_keep():
            return (self._number_of_dice * (self._number_to_keep_or_drop)) * len(
                self._single_dice_outcome_map.values()
            )
        return (
            self._number_of_dice * (self._number_of_dice - self._number_to_keep_or_drop)
        ) * len(self._single_dice_outcome_map.values())

    def get_probability_distribution(
        self,
    ) -> i_probability_distribution.IProbabilityDistribution:
        if self._simplified_form is not None:
            return self._simplified_form.get_probability_distribution()

        is_keep = self._is_keep()
        number_of_dice_to_select = self._number_of_dice - self._number_to_keep_or_drop
        number_of_dice_remaining = self._number_to_keep_or_drop
        if is_keep:
            number_of_dice_to_select = self._number_to_keep_or_drop
            number_of_dice_remaining = (
                self._number_of_dice - self._number_to_keep_or_drop
            )

        current = DropKeepExpression._build_dice_dict(
            number_of_dice_to_select, self._single_dice_outcome_map
        )
        current = DropKeepExpression._compute_iterations(
            current, number_of_dice_remaining, self._single_dice_outcome_map, is_keep
        )

        out_come_map = DropKeepExpression._collapse_outcomes(current)
        return probability_distribution.ProbabilityDistribution(out_come_map)

    @staticmethod
    def _build_dice_dict(
        number_of_dice: int, dice_outcome_map: typing.Dict[int, int]
    ) -> typing.Dict[str, int]:
        def safe_add_to_dict(
            dictionary: typing.Dict[str, int], key: str, value: int
        ) -> None:
            if key not in dictionary:
                dictionary[key] = 0
            dictionary[key] += value

        current_dict = {"": 1}
        for _ in range(number_of_dice):
            new_dict = {}
            for dice_outcome, count in dice_outcome_map.items():
                for old_key, old_value in current_dict.items():
                    new_key = DropKeepExpression._string_key_to_list(old_key)
                    new_key.append(dice_outcome)
                    new_key.sort()
                    safe_add_to_dict(
                        new_dict,
                        DropKeepExpression._int_list_to_string(new_key),
                        old_value * count,
                    )
            current_dict = new_dict
        return current_dict

    @staticmethod
    def _string_key_to_list(string: str) -> typing.List[int]:
        if string == "":
            return []
        return [int(n) for n in string.split("|")]

    @staticmethod
    def _int_list_to_string(values: typing.List[int]) -> str:
        values.sort()
        return "|".join([str(n) for n in values])

    @staticmethod
    def _update_key_list(old_key_string: str, new_value: int, is_keep: bool) -> str:
        old_value = DropKeepExpression._string_key_to_list(old_key_string)
        old_value.append(new_value)
        old_value.sort()
        if is_keep:
            old_value = old_value[1:]
        else:
            old_value = old_value[:-1]
        return DropKeepExpression._int_list_to_string(old_value)

    @staticmethod
    def _compute_iterations(
        current: typing.Dict[str, int],
        number_of_dice: int,
        dice_outcome_map: typing.Dict[int, int],
        is_keep: bool,
    ) -> typing.Dict[str, int]:
        def safe_add_to_dict(
            dictionary: typing.Dict[str, int], key: str, value: int
        ) -> None:
            if key not in dictionary:
                dictionary[key] = 0
            dictionary[key] += value

        current_dict = current
        for _ in range(number_of_dice):
            new_dict = {}
            for dice_outcome, count in dice_outcome_map.items():
                for old_key, old_value in current_dict.items():
                    new_key = DropKeepExpression._update_key_list(
                        old_key, dice_outcome, is_keep
                    )
                    safe_add_to_dict(new_dict, new_key, old_value * count)
            current_dict = new_dict
        return current_dict

    @staticmethod
    def _collapse_outcomes(outcomes: typing.Dict[str, int]) -> typing.Dict[int, int]:
        def safe_add_to_dict(
            dictionary: typing.Dict[int, int], key: int, value: int
        ) -> None:
            if key not in dictionary:
                dictionary[key] = 0
            dictionary[key] += value

        new_dict = {}
        for current_key, current_value in outcomes.items():
            total = sum(DropKeepExpression._string_key_to_list(current_key))
            safe_add_to_dict(new_dict, total, current_value)
        return new_dict



_NUMBER_OF_DICE = r"\d*"
_DICE_SYMBOL = r"d"

_DICE_NUM_TYPE = r"\d+"
_PERCENTILE_TYPE = r"%"
_FATE_TYPE = r"F"
_CUSTOM_TYPE_ENTRY = r"(-?\d+--?\d+(\*\d+)?|-?\d+(\*\d+)?)"
_CUSTOM_TYPE = (
    rf"\[(\s*{_CUSTOM_TYPE_ENTRY}\s*,\s*)*\s*{_CUSTOM_TYPE_ENTRY}\s*(,?)\s*\]"
)

_DICE_TYPE = f"({_DICE_NUM_TYPE}|{_PERCENTILE_TYPE}|{_FATE_TYPE}|{_CUSTOM_TYPE})"
_DROP_KEEP = r"[kd]\d+"

DICE_SYNTAX = _NUMBER_OF_DICE + _DICE_SYMBOL + _DICE_TYPE
DROP_KEEP_DICE_SYNTAX = DICE_SYNTAX + _DROP_KEEP



import copy
import io
import operator
import typing

import matplotlib.pyplot as pyplot
import PIL.Image as Image

import python_dice.interface.i_probability_distribution as i_probability_distribution


class ProbabilityDistribution(i_probability_distribution.IProbabilityDistribution):
    def __init__(self, result_map: typing.Optional[typing.Dict[int, int]] = None):
        self._outcome_count = 0
        self._result_map: typing.Dict[int, int] = (
            {} if result_map is None else result_map
        )
        for value in self._result_map.values():
            self._outcome_count += value

    def get_histogram(self) -> Image:
        histogram_data = self._get_histogram_form(self.get_dict_form())
        x_values = list(histogram_data.keys())
        return self._make_histogram(
            x_values, [histogram_data[x_value] for x_value in x_values], self.average()
        )

    def get_at_least_histogram(self) -> Image:
        histogram_data = self.at_least()
        x_values = list(histogram_data.keys())
        return self._make_histogram(
            x_values, [histogram_data[x_value] for x_value in x_values], self.average()
        )

    def get_at_most_histogram(self) -> Image:
        histogram_data = self.at_most()
        x_values = list(histogram_data.keys())
        return self._make_histogram(
            x_values, [histogram_data[x_value] for x_value in x_values], self.average()
        )

    def get_compare_histogram(
        self, other_probability: i_probability_distribution.IProbabilityDistribution
    ) -> Image:
        this_histogram_data = self._get_histogram_form(self.get_dict_form())
        other_histogram_data = self._get_histogram_form(
            other_probability.get_dict_form()
        )
        x_values = list(this_histogram_data.keys())
        x_values.extend(other_histogram_data.keys())

        def y_value_calculator(key):
            value = 0
            if key in this_histogram_data:
                value += this_histogram_data[key]
            if key in other_histogram_data:
                value -= other_histogram_data[key]
            return value

        return self._make_histogram(
            x_values, [y_value_calculator(x_value) for x_value in x_values]
        )

    def get_compare(
        self,
        other_probability: i_probability_distribution.IProbabilityDistribution,
        this_distribution_name: str = "This distribution",
        other_distribution_name: str = "other distribution",
    ) -> Image:
        this_at_least_data = self._get_histogram_form(self.get_dict_form())
        other_at_least_data = self._get_histogram_form(
            other_probability.get_dict_form()
        )

        x_values_one = list(this_at_least_data.keys())
        x_values_two = list(other_at_least_data.keys())
        return self._make_line_plot(
            x_values_one,
            [this_at_least_data[x_value] for x_value in x_values_one],
            x_values_two,
            [other_at_least_data[x_value] for x_value in x_values_two],
            this_distribution_name,
            other_distribution_name,
        )

    def get_compare_at_least(
        self,
        other_probability: i_probability_distribution.IProbabilityDistribution,
        this_distribution_name: str = "This distribution",
        other_distribution_name: str = "other distribution",
    ) -> Image:
        this_at_least_data = self.at_least()
        other_at_least_data = other_probability.at_least()

        x_values_one = list(this_at_least_data.keys())
        x_values_two = list(other_at_least_data.keys())
        return self._make_line_plot(
            x_values_one,
            [this_at_least_data[x_value] for x_value in x_values_one],
            x_values_two,
            [other_at_least_data[x_value] for x_value in x_values_two],
            this_distribution_name,
            other_distribution_name,
        )

    def get_compare_at_most(
        self,
        other_probability: i_probability_distribution.IProbabilityDistribution,
        this_distribution_name: str = "This distribution",
        other_distribution_name: str = "other distribution",
    ) -> Image:
        this_at_most_data = self.at_most()
        other_at_most_data = other_probability.at_most()

        x_values_one = list(this_at_most_data.keys())
        x_values_two = list(other_at_most_data.keys())
        return self._make_line_plot(
            x_values_one,
            [this_at_most_data[x_value] for x_value in x_values_one],
            x_values_two,
            [other_at_most_data[x_value] for x_value in x_values_two],
            this_distribution_name,
            other_distribution_name,
        )

    @staticmethod
    def _make_histogram(
        x_values: typing.List[float],
        y_values: typing.List[float],
        average: float = None,
    ) -> Image:
        x_values, y_values = (list(t) for t in zip(*sorted(zip(x_values, y_values))))
        bins = [value - 0.5 for value in x_values]
        bins.append(x_values[-1] + 0.5)

        _, axis = pyplot.subplots()
        axis.hist(x_values, weights=y_values, bins=bins)
        axis.grid(True, linestyle="-.")
        axis.set_ylabel("odds")
        axis.set_xlabel("outcome")
        if average is not None:
            max_height = max(y_values)
            axis.plot(
                [average, average],
                [0, max_height * 1.1],
                "-r",
                lw=2,
                label="Average = %d" % average,
            )
        buffer = io.BytesIO()
        pyplot.savefig(buffer, format="tiff")
        buffer.seek(0)
        image = Image.open(buffer)
        return image

    @staticmethod
    def _make_line_plot(
        x_values_one: typing.List[float],
        y_values_one: typing.List[float],
        x_values_two: typing.List[float],
        y_values_two: typing.List[float],
        lable_one: str = "Distribution One",
        lable_two: str = "Distribution Two",
    ) -> Image:
        x_values_one, y_values_one = (
            list(t) for t in zip(*sorted(zip(x_values_one, y_values_one)))
        )
        x_values_two, y_values_two = (
            list(t) for t in zip(*sorted(zip(x_values_two, y_values_two)))
        )
        _, axis = pyplot.subplots()
        axis.plot(x_values_one, y_values_one, "b", label=lable_one)
        axis.plot(x_values_two, y_values_two, "r", label=lable_two)
        axis.grid(True, linestyle="-.")
        axis.set_ylabel("odds")
        axis.set_xlabel("outcome")
        axis.legend()
        buffer = io.BytesIO()
        pyplot.savefig(buffer, format="png")
        buffer.seek(0)
        image = Image.open(buffer)
        return image

    def max(self) -> int:
        return max(self._result_map.keys())

    def min(self) -> int:
        return min(self._result_map.keys())

    def contains_zero(self) -> bool:
        return 0 in self._result_map and self._result_map[0] != 0

    def average(self) -> float:
        total_values = 0
        for item, value in self._result_map.items():
            total_values += item * value
        return total_values / self._outcome_count

    def at_least(self) -> typing.Dict[int, float]:
        at_least_dict = {}
        total_above = 1
        histogram_form = self._get_histogram_form(self.get_dict_form())
        keys = list(histogram_form.keys())
        keys = sorted(keys)
        for value in keys:
            at_least_dict[value] = total_above / 1
            total_above -= histogram_form[value]
        return at_least_dict

    def at_most(self) -> typing.Dict[int, float]:
        at_most_dict = {}
        at_or_below = 0
        histogram_form = self._get_histogram_form(self.get_dict_form())
        keys = list(histogram_form.keys())
        keys = sorted(keys)
        for value in keys:
            at_or_below += histogram_form[value]
            at_most_dict[value] = at_or_below / 1
        return at_most_dict

    def get_result_map(self) -> typing.Dict[int, int]:
        return self._result_map

    def get_dict_form(self) -> typing.Dict[int, float]:
        return {
            key: value / self._outcome_count for key, value in self._result_map.items()
        }

    @staticmethod
    def _get_histogram_form(
        base_data: typing.Dict[int, float]
    ) -> typing.Dict[int, float]:
        base_data = copy.copy(base_data)
        item_list = [(key, base_data[key]) for key in base_data.keys()]
        item_list.sort(key=lambda tup: tup[0])
        for int_value in range(item_list[0][0], item_list[-1][0] + 1):
            if int_value not in base_data:
                base_data[int_value] = 0
        return base_data

    def _combine_distributions(
        self,
        combination_function: typing.Callable[[int, int], int],
        other: i_probability_distribution.IProbabilityDistribution,
    ) -> typing.Dict[int, int]:
        new_result_map = {}

        def safe_add(key, value):
            if key not in new_result_map:
                new_result_map[key] = 0
            new_result_map[key] += value

        for this_key, this_value in self._result_map.items():
            for other_key, other_value in other.get_result_map().items():
                safe_add(
                    combination_function(this_key, other_key), this_value * other_value
                )
        return new_result_map

    def __add__(
        self, other: i_probability_distribution.IProbabilityDistribution
    ) -> "ProbabilityDistribution":
        new_result_map = self._combine_distributions(operator.add, other)
        return ProbabilityDistribution(new_result_map)

    def __sub__(
        self, other: i_probability_distribution.IProbabilityDistribution
    ) -> "ProbabilityDistribution":
        new_result_map = self._combine_distributions(operator.sub, other)
        return ProbabilityDistribution(new_result_map)

    def __mul__(
        self, other: i_probability_distribution.IProbabilityDistribution
    ) -> "ProbabilityDistribution":
        new_result_map = self._combine_distributions(operator.mul, other)
        return ProbabilityDistribution(new_result_map)

    def __floordiv__(
        self, other: i_probability_distribution.IProbabilityDistribution
    ) -> "ProbabilityDistribution":
        new_result_map = self._combine_distributions(operator.floordiv, other)
        return ProbabilityDistribution(new_result_map)

    def __eq__(
        self, other: i_probability_distribution.IProbabilityDistribution
    ) -> "ProbabilityDistribution":
        new_result_map = self._combine_distributions(
            lambda a, b: 1 if operator.eq(a, b) else 0, other
        )
        return ProbabilityDistribution(new_result_map)

    def __ne__(
        self, other: i_probability_distribution.IProbabilityDistribution
    ) -> "ProbabilityDistribution":
        new_result_map = self._combine_distributions(
            lambda a, b: 1 if operator.ne(a, b) else 0, other
        )
        return ProbabilityDistribution(new_result_map)

    def __lt__(
        self, other: i_probability_distribution.IProbabilityDistribution
    ) -> "ProbabilityDistribution":
        new_result_map = self._combine_distributions(
            lambda a, b: 1 if operator.lt(a, b) else 0, other
        )
        return ProbabilityDistribution(new_result_map)

    def __le__(
        self, other: i_probability_distribution.IProbabilityDistribution
    ) -> "ProbabilityDistribution":
        new_result_map = self._combine_distributions(
            lambda a, b: 1 if operator.le(a, b) else 0, other
        )
        return ProbabilityDistribution(new_result_map)

    def __gt__(
        self, other: i_probability_distribution.IProbabilityDistribution
    ) -> "ProbabilityDistribution":
        new_result_map = self._combine_distributions(
            lambda a, b: 1 if operator.gt(a, b) else 0, other
        )
        return ProbabilityDistribution(new_result_map)

    def __ge__(
        self, other: i_probability_distribution.IProbabilityDistribution
    ) -> "ProbabilityDistribution":
        new_result_map = self._combine_distributions(
            lambda a, b: 1 if operator.ge(a, b) else 0, other
        )
        return ProbabilityDistribution(new_result_map)

    def __and__(
        self, other: i_probability_distribution.IProbabilityDistribution
    ) -> "ProbabilityDistribution":
        new_result_map = self._combine_distributions(
            lambda a, b: 1
            if operator.and_(
                self._value_to_binary_value(a), self._value_to_binary_value(b)
            )
            else 0,
            other,
        )
        return ProbabilityDistribution(new_result_map)

    def __or__(
        self, other: i_probability_distribution.IProbabilityDistribution
    ) -> "ProbabilityDistribution":
        new_result_map = self._combine_distributions(
            lambda a, b: 1
            if operator.or_(
                self._value_to_binary_value(a), self._value_to_binary_value(b)
            )
            else 0,
            other,
        )
        return ProbabilityDistribution(new_result_map)

    @staticmethod
    def _value_to_binary_value(value: int):
        return value > 0

    def not_operator(self) -> "ProbabilityDistribution":
        new_result_map = {}

        def safe_add(key, value):
            if key not in new_result_map:
                new_result_map[key] = 0
            new_result_map[key] += value

        for result_key, result_value in self._result_map.items():
            if self._value_to_binary_value(result_key):
                safe_add(0, result_value)
            else:
                safe_add(1, result_value)
        return ProbabilityDistribution(new_result_map)

    def max_operator(
        self, other: "ProbabilityDistribution"
    ) -> "ProbabilityDistribution":
        new_result_map = {}

        def safe_add(key, value):
            if key not in new_result_map:
                new_result_map[key] = 0
            new_result_map[key] += value

        for result_key, result_value in self._result_map.items():
            for other_result_key, other_result_value in other.get_result_map().items():
                safe_add(
                    max(result_key, other_result_key), result_value * other_result_value
                )
        return ProbabilityDistribution(new_result_map)

    def min_operator(
        self, other: "ProbabilityDistribution"
    ) -> "ProbabilityDistribution":
        new_result_map = {}

        def safe_add(key, value):
            if key not in new_result_map:
                new_result_map[key] = 0
            new_result_map[key] += value

        for result_key, result_value in self._result_map.items():
            for other_result_key, other_result_value in other.get_result_map().items():
                safe_add(
                    min(result_key, other_result_key), result_value * other_result_value
                )
        return ProbabilityDistribution(new_result_map)

    def abs_operator(self) -> "ProbabilityDistribution":
        new_result_map = {}

        def safe_add(key, value):
            if key not in new_result_map:
                new_result_map[key] = 0
            new_result_map[key] += value

        for result_key, result_value in self._result_map.items():
            safe_add(abs(result_key), result_value)
        return ProbabilityDistribution(new_result_map)



import math
import typing

import rply

import python_dice.interface.i_probability_distribution as i_probability_distribution
import python_dice.interface.python_dice_expression.i_dice_expression as i_dice_expression


class NotExpression(i_dice_expression.IDiceExpression):

    TOKEN_RULE = """expression : NOT expression"""

    @staticmethod
    def add_production_function(
        parser_generator: rply.ParserGenerator,
    ) -> typing.Callable:
        @parser_generator.production(NotExpression.TOKEN_RULE)
        def not_operation(_, tokens) -> i_dice_expression.IDiceExpression:
            return NotExpression(tokens[1])

        return not_operation

    def __init__(self, expression: i_dice_expression.IDiceExpression):
        self._expression = expression

    def roll(self) -> int:
        return 0 if self._expression.roll() else 1

    def max(self) -> int:
        return self.get_probability_distribution().max()

    def min(self) -> int:
        return self.get_probability_distribution().min()

    def __str__(self) -> str:
        return f"!{str(self._expression)}"

    def estimated_cost(self) -> int:
        return int(math.floor(self._expression.estimated_cost() * 1.1))

    def get_probability_distribution(
        self,
    ) -> i_probability_distribution.IProbabilityDistribution:
        return self._expression.get_probability_distribution().not_operator()



import typing

import PIL.Image as Image

import python_dice.interface.i_probability_distribution as i_probability_distribution
import python_dice.interface.i_probability_state as i_probability_state
import python_dice.interface.i_python_dice_interpreter as i_python_dice_interpreter
import python_dice.interface.i_python_dice_parser as i_python_dice_parser
import python_dice.src.probability_distribution as probability_distribution
import python_dice.src.probability_state as probability_state
import python_dice.src.python_dice_parser as python_dice_parser


class PythonDiceInterpreter(i_python_dice_interpreter.IPythonDiceInterpreter):
    def __init__(
        self,
        parser: i_python_dice_parser.IPythonDiceParser = None,
        starting_state: i_probability_state.IProbabilityDistributionState = None,
    ):
        if parser is None:
            parser = python_dice_parser.PythonDiceParser()
        if starting_state is None:
            starting_state = probability_state.ProbabilityState()
        self._parser = parser
        self._state = starting_state

    def roll(self, input_text: typing.List[str]) -> typing.Dict[str, int]:
        return_dict = {}
        for line in input_text:
            token, _ = self._parser.parse(line, state=self._state)
            stdout = token.roll()
            return_dict = self._state.get_constant_dict()
            return_dict["stdout"] = stdout
        return return_dict

    def max(self, input_text: typing.List[str]) -> typing.Dict[str, int]:
        return_dict = {}
        for line in input_text:
            token, _ = self._parser.parse(line, state=self._state)
            stdout = token.max()
            return_dict = self._state.get_constant_dict()
            return_dict["stdout"] = stdout
        return return_dict

    def min(self, input_text: typing.List[str]) -> typing.Dict[str, int]:
        return_dict = {}
        for line in input_text:
            token, _ = self._parser.parse(line, state=self._state)
            stdout = token.min()
            return_dict = self._state.get_constant_dict()
            return_dict["stdout"] = stdout
        return return_dict

    def get_probability_distributions_dict(
        self, input_text: typing.List[str]
    ) -> typing.Dict[str, typing.Dict[int, float]]:
        return_dict = {}
        for line in input_text:
            token, _ = self._parser.parse(line, state=self._state)
            stdout = token.get_probability_distribution()
            return_dict = self._state.get_var_dict()
            return_dict["stdout"] = stdout
        return {key: value.get_dict_form() for key, value in return_dict.items()}

    def get_probability_distributions(
        self, input_text: typing.List[str]
    ) -> typing.Dict[str, i_probability_distribution.IProbabilityDistribution]:
        return_dict = {}
        for line in input_text:
            token, _ = self._parser.parse(line, state=self._state)
            stdout = token.get_probability_distribution()
            return_dict = self._state.get_var_dict()
            return_dict["stdout"] = stdout
        return {key: return_dict[key] for key in return_dict.keys()}

    def get_average(self, input_text: typing.List[str]) -> typing.Dict[str, float]:
        return_dict = {}
        for line in input_text:
            token, _ = self._parser.parse(line, state=self._state)
            stdout = token.get_probability_distribution()
            return_dict = self._state.get_var_dict()
            return_dict["stdout"] = stdout
        return {key: value.average() for key, value in return_dict.items()}

    def get_histogram(self, input_text: typing.List[str]) -> Image:
        stdout = probability_distribution.ProbabilityDistribution()
        for line in input_text:
            token, _ = self._parser.parse(line, state=self._state)
            stdout = token.get_probability_distribution()
        return stdout.get_histogram()

    def get_at_least_histogram(self, input_text: typing.List[str]) -> Image:
        stdout = probability_distribution.ProbabilityDistribution()
        for line in input_text:
            token, _ = self._parser.parse(line, state=self._state)
            stdout = token.get_probability_distribution()
        return stdout.get_at_least_histogram()

    def get_at_most_histogram(self, input_text: typing.List[str]) -> Image:
        stdout = probability_distribution.ProbabilityDistribution()
        for line in input_text:
            token, _ = self._parser.parse(line, state=self._state)
            stdout = token.get_probability_distribution()
        return stdout.get_at_most_histogram()

    def get_estimated_cost(self, input_text: typing.List[str]) -> int:
        total_cost = 0
        for line in input_text:
            token, _ = self._parser.parse(line, state=self._state)
            line_cost = token.estimated_cost()
            total_cost += line_cost
        return total_cost



import python_dice.interface.python_dice_syntax.i_dice_syntax as i_dice_statement


class AddSyntax(i_dice_statement.IDiceSyntax):
    TOKEN_NAME = "ADD"
    TOKEN_REGEX = r"\+"

    @staticmethod
    def get_token_name() -> str:
        return AddSyntax.TOKEN_NAME

    @staticmethod
    def get_token_regex() -> str:
        return AddSyntax.TOKEN_REGEX



import typing

import rply

import python_dice.interface.i_probability_distribution as i_probability_distribution
import python_dice.interface.python_dice_expression.i_dice_expression as i_dice_expression
import python_dice.src.probability_state as probability_state


class VarAssignmentExpression(i_dice_expression.IDiceExpression):

    TOKEN_RULE = """expression : VAR NAME ASSIGNMENT expression"""

    @staticmethod
    def add_production_function(
        parser_generator: rply.ParserGenerator,
    ) -> typing.Callable:
        @parser_generator.production(VarAssignmentExpression.TOKEN_RULE)
        def var_assignment_operation(
            state, tokens
        ) -> i_dice_expression.IDiceExpression:
            return VarAssignmentExpression(state, tokens[1].value, tokens[3])

        return var_assignment_operation

    def __init__(
        self,
        state: probability_state.ProbabilityState,
        name: str,
        expression: i_dice_expression.IDiceExpression,
    ):
        self._state = state
        self._name = name
        self._expression = expression

    def roll(self) -> int:
        roll_value = self._expression.roll()
        self._state.set_constant(self._name, roll_value)
        return roll_value

    def max(self) -> int:
        max_value = self._expression.max()
        self._state.set_constant(self._name, max_value)
        self._state.set_var(self._name, self._expression.get_probability_distribution())
        return max_value

    def min(self) -> int:
        min_value = self._expression.min()
        self._state.set_constant(self._name, min_value)
        self._state.set_var(self._name, self._expression.get_probability_distribution())
        return min_value

    def __str__(self) -> str:
        return f"VAR {self._name} = {str(self._expression)}"

    def estimated_cost(self) -> int:
        cost = self._expression.estimated_cost()
        self._state.set_constant(self._name, cost)
        return 2 + cost

    def get_probability_distribution(
        self,
    ) -> i_probability_distribution.IProbabilityDistribution:
        distribution = self._expression.get_probability_distribution()
        self._state.set_var(self._name, distribution)
        return distribution



import typing

import rply

import python_dice.interface.i_python_dice_lexer as i_pydice_lexer
import python_dice.src.python_dice_syntax_objects as pydice_syntax_objects


class PythonDiceLexer(i_pydice_lexer.IPythonDiceLexer):
    def __init__(self):
        partial_lexer = rply.LexerGenerator()
        for syntax in pydice_syntax_objects.LEXER_SYNTAX:
            partial_lexer.add(syntax.get_token_name(), syntax.get_token_regex())
        # Ignore spaces
        partial_lexer.ignore(r"\s+")
        self._lexer = partial_lexer.build()

    def lex(self, input_text: str) -> typing.List[rply.Token]:
        try:
            return list(self._lexer.lex(input_text))
        except rply.errors.LexingError as exception:
            input_lines = input_text.split("\n")
            source_column_number = exception.getsourcepos().idx

            current_line_index = 0
            while len(input_lines[current_line_index]) < source_column_number:
                source_column_number -= len(input_lines[current_line_index]) + 1
                current_line_index += 1

            sub_string_start = max(0, source_column_number - 50)
            sub_string_end = min(len(input_text), source_column_number + 50)

            sub_string = input_lines[current_line_index][
                sub_string_start:sub_string_end
            ]
            arrow_line = "-" * (source_column_number - sub_string_start) + "^"
            message = f"{sub_string}\n{arrow_line}"

            new_exception = ValueError(message)
        raise new_exception



import abc
import typing

import python_dice.interface.i_probability_state as i_probability_state
import python_dice.interface.python_dice_expression.i_dice_expression as i_dice_expression


class IPythonDiceParser(abc.ABC):
    @abc.abstractmethod
    def parse(
        self,
        input_text: str,
        state: i_probability_state.IProbabilityDistributionState = None,
    ) -> typing.Tuple[
        i_dice_expression.IDiceExpression,
        i_probability_state.IProbabilityDistributionState,
    ]:
        """

        :return: output of pydice program
        """



import python_dice.interface.python_dice_syntax.i_dice_syntax as i_dice_statement


class MinMaxSyntax(i_dice_statement.IDiceSyntax):
    TOKEN_NAME = "MINMAX"
    TOKEN_REGEX = r"\bMAX\b|\bMIN\b"

    @staticmethod
    def get_token_name() -> str:
        return MinMaxSyntax.TOKEN_NAME

    @staticmethod
    def get_token_regex() -> str:
        return MinMaxSyntax.TOKEN_REGEX



import setuptools

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

setuptools.setup(
    name="python_dice",
    version="1.4.5",
    author="Mark Brockett Robson",
    author_email="mark.brockett.robson@gmail.com",
    description="a statistical dice library for python",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/markbrockettrobson/python_dice",
    packages=[
        "python_dice",
        "python_dice.interface",
        "python_dice.interface.python_dice_expression",
        "python_dice.interface.python_dice_syntax",
        "python_dice.src",
        "python_dice.src.python_dice_expression",
        "python_dice.src.python_dice_syntax",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "matplotlib>=3.1.1",
        "rply>=0.7.7",
        "networkx>=2.3",
        "numpy>=1.17.3",
        "pillow>=7.2.0",
    ],
    python_requires=">=3.6",
)



import operator
import typing

import rply

import python_dice.interface.i_probability_distribution as i_probability_distribution
import python_dice.interface.python_dice_expression.i_dice_expression as i_dice_expression


class BinaryOperatorExpression(i_dice_expression.IDiceExpression):
    RULE = """expression : expression BINARY_OPERATOR expression"""
    OPERATOR_MAP = {
        "==": operator.eq,
        "!=": operator.ne,
        "<=": operator.le,
        "<": operator.lt,
        ">=": operator.ge,
        ">": operator.gt,
        "AND": operator.and_,
        "OR": operator.or_,
    }

    @staticmethod
    def add_production_function(
        parser_generator: rply.ParserGenerator,
    ) -> typing.Callable:
        @parser_generator.production(BinaryOperatorExpression.RULE)
        def binary_operator(_, tokens) -> i_dice_expression.IDiceExpression:
            return BinaryOperatorExpression(tokens[1].value, tokens[0], tokens[2])

        return binary_operator

    def __init__(
        self,
        operator_str: str,
        expression_one: i_dice_expression.IDiceExpression,
        expression_two: i_dice_expression.IDiceExpression,
    ):
        self._operator = operator_str
        self._expression_one = expression_one
        self._expression_two = expression_two

    def roll(self) -> int:
        return (
            1
            if self.OPERATOR_MAP[self._operator](
                self._expression_one.roll(), self._expression_two.roll()
            )
            else 0
        )

    def max(self) -> int:
        return self.get_probability_distribution().max()

    def min(self) -> int:
        return self.get_probability_distribution().min()

    def __str__(self) -> str:
        return (
            f"{str(self._expression_one)} {self._operator} {str(self._expression_two)}"
        )

    def estimated_cost(self) -> int:
        return (
            self._expression_one.estimated_cost()
            + self._expression_two.estimated_cost()
        )

    def get_probability_distribution(
        self,
    ) -> i_probability_distribution.IProbabilityDistribution:
        return self.OPERATOR_MAP[self._operator](
            self._expression_one.get_probability_distribution(),
            self._expression_two.get_probability_distribution(),
        )



import abc

import rply


class IDiceSyntax(abc.ABC, rply.token.BaseBox):
    @staticmethod
    @abc.abstractmethod
    def get_token_name() -> str:
        """

        :return: the token name for this syntax
                 <str>
        """

    @staticmethod
    @abc.abstractmethod
    def get_token_regex() -> str:
        """

        :return: the token regex for this syntax
                 <str>
        """



import abc
import typing

import PIL.Image as Image


class IProbabilityDistribution(abc.ABC):
    @abc.abstractmethod
    def get_dict_form(self) -> typing.Dict[int, float]:
        """

        :return: a dict mapping an outcome of the program to a probability of that outcome
                <typing.Dict[int, float]>
        """

    @abc.abstractmethod
    def get_result_map(self) -> typing.Dict[int, int]:
        """

        :return: a dict mapping an outcome of the program to the number of ways to achieve that outcome
                 <typing.Dict[int, int]>
        """

    @abc.abstractmethod
    def get_histogram(self) -> Image:
        """

        prints a plot of the distribution with matplotlib
        :returns the pyplot figure for plot
                 <Image>
        """

    @abc.abstractmethod
    def get_at_least_histogram(self) -> Image:
        """

        prints a plot of the area under the distribution with matplotlib
        :returns the pyplot figure for plot
                 <Image>
        """

    @abc.abstractmethod
    def get_at_most_histogram(self) -> Image:
        """

        prints a plot of 1 - the area under the distribution with matplotlib
        :returns the pyplot figure for plot
                 <Image>
        """

    @abc.abstractmethod
    def get_compare_histogram(
        self, other_probability: "IProbabilityDistribution"
    ) -> Image:
        """

        prints a plot of this distribution - other distribution with matplotlib
        :returns the pyplot figure for plot
                 <Image>
        """

    @abc.abstractmethod
    def get_compare(
        self,
        other_probability: "IProbabilityDistribution",
        this_distribution_name: str,
        other_distribution_name: str,
    ) -> Image:
        """

        prints a plot of this at least this distribution alongside another distribution with matplotlib
        :returns the pyplot figure for plot
                 <Image>
        """

    @abc.abstractmethod
    def get_compare_at_least(
        self,
        other_probability: "IProbabilityDistribution",
        this_distribution_name: str,
        other_distribution_name: str,
    ) -> Image:
        """

        prints a plot of this at least this distribution alongside another distribution with matplotlib
        :returns the pyplot figure for plot
                 <Image>
        """

    @abc.abstractmethod
    def get_compare_at_most(
        self,
        other_probability: "IProbabilityDistribution",
        this_distribution_name: str,
        other_distribution_name: str,
    ) -> Image:
        """

        prints a plot of this at most this distribution alongside another distribution with matplotlib
        :returns the pyplot figure for plot
                 <Image>
        """

    @abc.abstractmethod
    def max(self) -> int:
        """

        :returns: the max roll that could be made
                  <int>
        """

    @abc.abstractmethod
    def min(self) -> int:
        """

        :returns: the min roll that could be made
                  <int>
        """

    @abc.abstractmethod
    def contains_zero(self) -> bool:
        """

        :returns: returns true if the distribution contains zero
                  <bool>
        """

    @abc.abstractmethod
    def average(self) -> float:
        """

        :return: returns the waited average of the distribution
                 <float>
        """

    @abc.abstractmethod
    def at_least(self) -> typing.Dict[int, float]:
        pass

    @abc.abstractmethod
    def at_most(self) -> typing.Dict[int, float]:
        pass

    @abc.abstractmethod
    def __add__(self, other: "IProbabilityDistribution") -> "IProbabilityDistribution":
        pass

    @abc.abstractmethod
    def __sub__(self, other: "IProbabilityDistribution") -> "IProbabilityDistribution":
        pass

    @abc.abstractmethod
    def __mul__(self, other: "IProbabilityDistribution") -> "IProbabilityDistribution":
        pass

    @abc.abstractmethod
    def __floordiv__(
        self, other: "IProbabilityDistribution"
    ) -> "IProbabilityDistribution":
        pass

    @abc.abstractmethod
    def __eq__(self, other: "IProbabilityDistribution") -> "IProbabilityDistribution":
        pass

    @abc.abstractmethod
    def __ne__(self, other: "IProbabilityDistribution") -> "IProbabilityDistribution":
        pass

    @abc.abstractmethod
    def __lt__(self, other: "IProbabilityDistribution") -> "IProbabilityDistribution":
        pass

    @abc.abstractmethod
    def __le__(self, other: "IProbabilityDistribution") -> "IProbabilityDistribution":
        pass

    @abc.abstractmethod
    def __gt__(self, other: "IProbabilityDistribution") -> "IProbabilityDistribution":
        pass

    @abc.abstractmethod
    def __ge__(self, other: "IProbabilityDistribution") -> "IProbabilityDistribution":
        pass

    @abc.abstractmethod
    def __and__(self, other: "IProbabilityDistribution") -> "IProbabilityDistribution":
        pass

    @abc.abstractmethod
    def __or__(self, other: "IProbabilityDistribution") -> "IProbabilityDistribution":
        pass

    @abc.abstractmethod
    def not_operator(self) -> "IProbabilityDistribution":
        pass

    @abc.abstractmethod
    def max_operator(
        self, other: "IProbabilityDistribution"
    ) -> "IProbabilityDistribution":
        pass

    @abc.abstractmethod
    def min_operator(
        self, other: "IProbabilityDistribution"
    ) -> "IProbabilityDistribution":
        pass

    @abc.abstractmethod
    def abs_operator(self) -> "IProbabilityDistribution":
        pass



import abc
import typing

import PIL.Image as Image

import python_dice.interface.i_probability_distribution as i_probability_distribution


class IPythonDiceInterpreter(abc.ABC):
    @abc.abstractmethod
    def roll(self, input_text: typing.List[str]) -> typing.Dict[str, int]:
        """

        :return: output of state of the python program
        """

    @abc.abstractmethod
    def max(self, input_text: typing.List[str]) -> typing.Dict[str, int]:
        """

        :return: output of state of the python program
        """

    @abc.abstractmethod
    def min(self, input_text: typing.List[str]) -> typing.Dict[str, int]:
        """

        :return: output of state of the python program
        """

    @abc.abstractmethod
    def get_probability_distributions_dict(
        self, input_text: typing.List[str]
    ) -> typing.Dict[str, typing.Dict[int, float]]:
        """

        :return: output of state of the python program
        """

    @abc.abstractmethod
    def get_probability_distributions(
        self, input_text: typing.List[str]
    ) -> typing.Dict[str, i_probability_distribution.IProbabilityDistribution]:
        """

        :return: output of state of the python program
        """

    @abc.abstractmethod
    def get_histogram(self, input_text: typing.List[str]) -> Image:
        """

        :return: output of state of the python program
        """

    @abc.abstractmethod
    def get_at_least_histogram(self, input_text: typing.List[str]) -> Image:
        """

        :return: output of state of the python program
        """

    @abc.abstractmethod
    def get_at_most_histogram(self, input_text: typing.List[str]) -> Image:
        """

        :return: output of state of the python program
        """

    @abc.abstractmethod
    def get_average(self, input_text: typing.List[str]) -> typing.Dict[str, float]:
        """

        :return: output of state of the python program
        """

    @abc.abstractmethod
    def get_estimated_cost(self, input_text: typing.List[str]) -> int:
        """

        :return: total cost estimate for python dice program
        """
